# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import logging

import wrapt

from scout_apm.core.tracked_request import TrackedRequest

try:
    import redis
except ImportError:
    redis = None
else:
    if redis.VERSION[0] >= 3:
        from redis import Redis
        from redis.client import Pipeline
    else:  # pragma: no cover
        from redis import StrictRedis as Redis
        from redis.client import BasePipeline as Pipeline

logger = logging.getLogger(__name__)


class Instrument(object):
    installed = False

    def installable(self):
        if redis is None:
            logger.info("Unable to import for Redis instruments")
            return False
        if self.installed:
            logger.warning("Redis Instruments are already installed.")
            return False
        return True

    def install(self):
        if not self.installable():
            logger.info("Redis instruments are not installable. Skipping.")
            return False

        self.__class__.installed = True

        self.patch_redis()
        self.patch_pipeline()

        logger.info("Instrumented Redis")
        return True

    def uninstall(self):
        if not self.installed:
            logger.info("Redis instruments are not installed. Skipping.")
            return False

        self.__class__.installed = False

        self.unpatch_redis()
        self.unpatch_pipeline()

    def patch_redis(self):
        @wrapt.decorator
        def wrapped_execute_command(wrapped, instance, args, kwargs):
            try:
                op = args[0]
            except (IndexError, TypeError):
                op = "Unknown"

            tracked_request = TrackedRequest.instance()
            tracked_request.start_span(operation="Redis/{}".format(op))

            try:
                return wrapped(*args, **kwargs)
            finally:
                tracked_request.stop_span()

        try:
            Redis.execute_command = wrapped_execute_command(Redis.execute_command)
        except Exception as exc:
            logger.warning(
                "Unable to instrument for Redis Redis.execute_command: %r",
                exc,
                exc_info=exc,
            )

    def unpatch_redis(self):
        Redis.execute_command = Redis.execute_command.__wrapped__

    def patch_pipeline(self):
        @wrapt.decorator
        def wrapped_execute(wrapped, instance, args, kwargs):
            tracked_request = TrackedRequest.instance()
            tracked_request.start_span(operation="Redis/MULTI")

            try:
                return wrapped(*args, **kwargs)
            finally:
                tracked_request.stop_span()

        try:
            Pipeline.execute = wrapped_execute(Pipeline.execute)
        except Exception as exc:
            logger.warning(
                "Unable to instrument for Redis BasePipeline.execute: %r",
                exc,
                exc_info=exc,
            )

    def unpatch_pipeline(self):
        Pipeline.execute = Pipeline.execute.__wrapped__



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import wrapt
from flask import current_app
from flask.globals import _request_ctx_stack

import scout_apm.core
from scout_apm.core.config import scout_config
from scout_apm.core.tracked_request import TrackedRequest
from scout_apm.core.web_requests import werkzeug_track_request_data


class ScoutApm(object):
    def __init__(self, app):
        self.app = app
        app.before_first_request(self.before_first_request)
        app.dispatch_request = self.wrapped_dispatch_request(app.dispatch_request)

    #############
    #  Startup  #
    #############

    def before_first_request(self):
        self.extract_flask_settings()
        installed = scout_apm.core.install()
        self._do_nothing = not installed

    def extract_flask_settings(self):
        """
        Copies SCOUT_* settings in the app into Scout's config lookup
        """
        configs = {}
        configs["application_root"] = self.app.instance_path
        for name in current_app.config:
            if name.startswith("SCOUT_"):
                value = current_app.config[name]
                clean_name = name.replace("SCOUT_", "").lower()
                configs[clean_name] = value
        scout_config.set(**configs)

    ############################
    #  Request Lifecycle hook  #
    ############################

    @wrapt.decorator
    def wrapped_dispatch_request(self, wrapped, instance, args, kwargs):
        if self._do_nothing:
            return wrapped(*args, **kwargs)

        request = _request_ctx_stack.top.request

        # Copied logic from Flask
        if request.routing_exception is not None:
            return wrapped(*args, **kwargs)

        rule = request.url_rule
        view_func = self.app.view_functions[rule.endpoint]

        name = view_func.__module__ + "." + view_func.__name__
        operation = "Controller/" + name

        tracked_request = TrackedRequest.instance()
        tracked_request.is_real_request = True
        span = tracked_request.start_span(
            operation=operation, should_capture_backtrace=False
        )
        span.tag("name", name)

        werkzeug_track_request_data(request, tracked_request)

        try:
            return wrapped(*args, **kwargs)
        except Exception as exc:
            tracked_request.tag("error", "true")
            raise exc
        finally:
            tracked_request.stop_span()



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import logging

import wrapt

from scout_apm.core.tracked_request import TrackedRequest

try:
    from jinja2 import Template
except ImportError:
    Template = None

logger = logging.getLogger(__name__)


class Instrument(object):
    installed = False

    def installable(self):
        if Template is None:
            logger.info("Unable to import for Jinja2 instruments")
            return False
        if self.installed:
            logger.warning("Jinja2 Instruments are already installed.")
            return False
        return True

    def install(self):
        if not self.installable():
            logger.info("Jinja2 instruments are not installable. Skipping.")
            return False

        self.__class__.installed = True

        try:
            Template.render = wrapped_render(Template.render)
        except Exception as exc:
            logger.warning(
                "Unable to instrument for Jinja2 Template.render: %r", exc, exc_info=exc
            )
            return False
        logger.info("Instrumented Jinja2")
        return True

    def uninstall(self):
        if not self.installed:
            logger.info("Jinja2 instruments are not installed. Skipping.")
            return False

        self.__class__.installed = False

        Template.render = Template.render.__wrapped__


@wrapt.decorator
def wrapped_render(wrapped, instance, args, kwargs):
    tracked_request = TrackedRequest.instance()
    span = tracked_request.start_span(operation="Template/Render")
    span.tag("name", instance.name)
    try:
        return wrapped(*args, **kwargs)
    finally:
        tracked_request.stop_span()



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import logging

import wrapt

from scout_apm.core.tracked_request import TrackedRequest

try:
    from elasticsearch import Elasticsearch, Transport
except ImportError:
    Elasticsearch = None
    Transport = None

logger = logging.getLogger(__name__)


class Instrument(object):
    installed = False

    CLIENT_METHODS = [
        "bulk",
        "count",
        "create",
        "delete",
        "delete_by_query",
        "exists",
        "exists_source",
        "explain",
        "field_caps",
        "get",
        "get_source",
        "index",
        "mget",
        "msearch",
        "msearch_template",
        "mtermvectors",
        "reindex",
        "reindex_rethrottle",
        "search",
        "search_shards",
        "search_template",
        "termvectors",
        "update",
        "update_by_query",
    ]

    def installable(self):
        if Elasticsearch is None:
            logger.info("Unable to import for Elasticsearch instruments")
            return False
        if self.installed:
            logger.warning("Elasticsearch Instruments are already installed.")
            return False
        return True

    def install(self):
        if not self.installable():
            logger.info("Elasticsearch instruments are not installable. Skipping.")
            return False

        self.__class__.installed = True

        self.instrument_client()
        self.instrument_transport()

    def uninstall(self):
        if not self.installed:
            logger.info("Elasticsearch instruments are not installed. Skipping.")
            return False

        self.uninstrument_client()
        self.uninstrument_transport()

        self.__class__.installed = False

    def instrument_client(self):
        for name in self.__class__.CLIENT_METHODS:
            method = getattr(Elasticsearch, name, None)
            if method is not None:
                setattr(Elasticsearch, name, wrap_client_method(method))

    def uninstrument_client(self):
        for name in self.__class__.CLIENT_METHODS:
            method = getattr(Elasticsearch, name, None)
            if method is not None:
                setattr(Elasticsearch, name, method.__wrapped__)

    def instrument_transport(self):
        try:

            def _sanitize_name(name):
                try:
                    op = name.split("/")[-1]
                    op = op[1:]  # chop leading '_' from op
                    allowed_names = [
                        "bench",
                        "bulk",
                        "count",
                        "exists",
                        "explain",
                        "field_stats",
                        "health",
                        "mget",
                        "mlt",
                        "mpercolate",
                        "msearch",
                        "mtermvectors",
                        "percolate",
                        "query",
                        "scroll",
                        "search_shards",
                        "source",
                        "suggest",
                        "template",
                        "termvectors",
                        "update",
                        "search",
                    ]

                    if op in allowed_names:
                        return op.title()
                    return "Unknown"
                except Exception:
                    return "Unknown"

            @wrapt.decorator
            def wrapped_perform_request(wrapped, instance, args, kwargs):
                try:
                    op = _sanitize_name(args[1])
                except IndexError:
                    op = "Unknown"

                tracked_request = TrackedRequest.instance()
                tracked_request.start_span(
                    operation="Elasticsearch/{}".format(op), ignore_children=True
                )

                try:
                    return wrapped(*args, **kwargs)
                finally:
                    tracked_request.stop_span()

            Transport.perform_request = wrapped_perform_request(
                Transport.perform_request
            )

            logger.info("Instrumented Elasticsearch Transport")

        except Exception as e:
            logger.warning(
                "Unable to instrument for Elasticsearch Transport.perform_request: %r",
                e,
            )
            return False
        return True

    def uninstrument_transport(self):

        Transport.perform_request = Transport.perform_request.__wrapped__


@wrapt.decorator
def wrap_client_method(wrapped, instance, args, kwargs):
    def _get_index(index, *args, **kwargs):
        return index

    try:
        index = _get_index(*args, **kwargs)
    except TypeError:
        index = "Unknown"
    else:
        if not index:
            index = "Unknown"
        if isinstance(index, (list, tuple)):
            index = ",".join(index)
    index = index.title()
    camel_name = "".join(c.title() for c in wrapped.__name__.split("_"))
    operation = "Elasticsearch/{}/{}".format(index, camel_name)
    tracked_request = TrackedRequest.instance()
    tracked_request.start_span(operation=operation, ignore_children=True)

    try:
        return wrapped(*args, **kwargs)
    finally:
        tracked_request.stop_span()



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import scout_apm.core
from scout_apm.compat import ContextDecorator, text
from scout_apm.core.config import ScoutConfig
from scout_apm.core.tracked_request import TrackedRequest

__all__ = [
    "BackgroundTransaction",
    "Config",
    "Context",
    "WebTransaction",
    "install",
    "instrument",
]


class Context(object):
    @classmethod
    def add(self, key, value):
        """Adds context to the currently executing request.

        :key: Any String identifying the request context.
              Example: "user_ip", "plan", "alert_count"
        :value: Any json-serializable type.
              Example: "1.1.1.1", "free", 100
        :returns: nothing.
        """
        TrackedRequest.instance().tag(key, value)


class Config(ScoutConfig):
    pass


install = scout_apm.core.install


def ignore_transaction():
    TrackedRequest.instance().tag("ignore_transaction", True)


class instrument(ContextDecorator):
    def __init__(self, operation, kind="Custom", tags=None):
        self.operation = text(kind) + "/" + text(operation)
        if tags is None:
            self.tags = {}
        else:
            self.tags = tags

    def __enter__(self):
        tracked_request = TrackedRequest.instance()
        self.span = tracked_request.start_span(operation=self.operation)
        for key, value in self.tags.items():
            self.tag(key, value)
        return self

    def __exit__(self, *exc):
        tracked_request = TrackedRequest.instance()
        tracked_request.stop_span()
        return False

    def tag(self, key, value):
        if self.span is not None:
            self.span.tag(key, value)


class Transaction(ContextDecorator):
    """
    This Class is not meant to be used directly.
    Use one of the subclasses
    (WebTransaction or BackgroundTransaction)
    """

    def __init__(self, name, tags=None):
        self.name = text(name)
        if tags is None:
            self.tags = {}
        else:
            self.tags = tags

    @classmethod
    def start(cls, kind, name, tags=None):
        operation = text(kind) + "/" + text(name)

        tracked_request = TrackedRequest.instance()
        tracked_request.is_real_request = True
        span = tracked_request.start_span(
            operation=operation, should_capture_backtrace=False
        )
        if tags is not None:
            for key, value in tags.items():
                tracked_request.tag(key, value)
        return span

    @classmethod
    def stop(cls):
        tracked_request = TrackedRequest.instance()
        tracked_request.stop_span()
        return True

    # __enter__ must be defined by child classes.

    # *exc is any exception raised. Ignore that
    def __exit__(self, *exc):
        WebTransaction.stop()
        return False

    def tag(self, key, value):
        if self.span is not None:
            self.span.tag(key, value)


class WebTransaction(Transaction):
    @classmethod
    def start(cls, name, tags=None):
        super(WebTransaction, cls).start("Controller", text(name), tags)

    def __enter__(self):
        super(WebTransaction, self).start("Controller", self.name, self.tags)


class BackgroundTransaction(Transaction):
    @classmethod
    def start(cls, name, tags=None):
        super(BackgroundTransaction, cls).start("Job", text(name), tags)

    def __enter__(self):
        super(BackgroundTransaction, self).start("Job", self.name, self.tags)


def rename_transaction(name):
    if name is not None:
        tracked_request = TrackedRequest.instance()
        tracked_request.tag("transaction.name", name)



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import logging

import wrapt

from scout_apm.compat import text_type
from scout_apm.core.tracked_request import TrackedRequest

try:
    from urllib3 import HTTPConnectionPool
except ImportError:
    HTTPConnectionPool = None

logger = logging.getLogger(__name__)


class Instrument(object):
    installed = False

    def installable(self):
        if HTTPConnectionPool is None:
            logger.info("Unable to import for Urllib3 instruments")
            return False
        if self.installed:
            logger.warning("Urllib3 Instruments are already installed.")
            return False
        return True

    def install(self):
        if not self.installable():
            logger.info("Urllib3 instruments are not installable. Skipping.")
            return False

        self.__class__.installed = True

        try:
            HTTPConnectionPool.urlopen = wrapped_urlopen(HTTPConnectionPool.urlopen)
        except Exception as exc:
            logger.warning(
                "Unable to instrument for Urllib3 HTTPConnectionPool.urlopen: %r",
                exc,
                exc_info=exc,
            )
            return False

        logger.info("Instrumented Urllib3")
        return True

    def uninstall(self):
        if not self.installed:
            logger.info("Urllib3 instruments are not installed. Skipping.")
            return False

        self.__class__.installed = False

        HTTPConnectionPool.urlopen = HTTPConnectionPool.urlopen.__wrapped__


@wrapt.decorator
def wrapped_urlopen(wrapped, instance, args, kwargs):
    def _extract_method(method, *args, **kwargs):
        return method

    try:
        method = _extract_method(*args, **kwargs)
    except TypeError:
        method = "Unknown"

    try:
        url = text_type(instance._absolute_url("/"))
    except Exception:
        logger.exception("Could not get URL for HTTPConnectionPool")
        url = "Unknown"

    tracked_request = TrackedRequest.instance()
    span = tracked_request.start_span(operation="HTTP/{}".format(method))
    span.tag("url", text_type(url))

    try:
        return wrapped(*args, **kwargs)
    finally:
        tracked_request.stop_span()



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import datetime as dt
import sys
from os import getpid

from scout_apm.core.commands import ApplicationEvent
from scout_apm.core.config import scout_config
from scout_apm.core.socket import CoreAgentSocket


def report_app_metadata():
    CoreAgentSocket.instance().send(
        ApplicationEvent(
            event_type="scout.metadata",
            event_value=get_metadata(),
            source="Pid: " + str(getpid()),
            timestamp=dt.datetime.utcnow(),
        )
    )


def get_metadata():
    data = {
        "language": "python",
        "language_version": "{}.{}.{}".format(*sys.version_info[:3]),
        "server_time": dt.datetime.utcnow().isoformat() + "Z",
        "framework": scout_config.value("framework"),
        "framework_version": scout_config.value("framework_version"),
        "environment": "",
        "app_server": scout_config.value("app_server"),
        "hostname": scout_config.value("hostname"),
        "database_engine": "",
        "database_adapter": "",
        "application_name": "",
        "libraries": get_python_packages_versions(),
        "paas": "",
        "application_root": scout_config.value("application_root"),
        "scm_subdirectory": scout_config.value("scm_subdirectory"),
        "git_sha": scout_config.value("revision_sha"),
    }
    # Deprecated - see #327:
    data["version"] = data["language_version"]
    return data


def get_python_packages_versions():
    try:
        import pkg_resources
    except ImportError:
        return []

    return list(
        sorted(
            (distribution.project_name, distribution.version)
            for distribution in pkg_resources.working_set
        )
    )



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import logging

import wrapt

from scout_apm.core.tracked_request import TrackedRequest

try:
    from pymongo.collection import Collection
except ImportError:
    Collection = None

logger = logging.getLogger(__name__)


class Instrument(object):
    installed = False

    PYMONGO_METHODS = [
        "aggregate",
        "bulk_write",
        "count",
        "create_index",
        "create_indexes",
        "delete_many",
        "delete_one",
        "distinct",
        "drop",
        "drop_index",
        "drop_indexes",
        "ensure_index",
        "find_and_modify",
        "find_one",
        "find_one_and_delete",
        "find_one_and_replace",
        "find_one_and_update",
        "group",
        "inline_map_reduce",
        "insert",
        "insert_many",
        "insert_one",
        "map_reduce",
        "reindex",
        "remove",
        "rename",
        "replace_one",
        "save",
        "update",
        "update_many",
        "update_one",
    ]

    def installable(self):
        if Collection is None:
            logger.info("Unable to import for PyMongo instruments")
            return False
        if self.installed:
            logger.warning("PyMongo Instruments are already installed.")
            return False
        return True

    def install(self):
        if not self.installable():
            logger.info("PyMongo instruments are not installable. Skipping.")
            return False

        self.__class__.installed = True

        for name in self.__class__.PYMONGO_METHODS:
            method = getattr(Collection, name, None)
            if method is not None:
                setattr(Collection, name, wrap_collection_method(method))
        return True

    def uninstall(self):
        if not self.installed:
            logger.info("PyMongo instruments are not installed. Skipping.")
            return False

        for name in self.__class__.PYMONGO_METHODS:
            method = getattr(Collection, name, None)
            if method is not None:
                setattr(Collection, name, method.__wrapped__)

        self.__class__.installed = False


@wrapt.decorator
def wrap_collection_method(wrapped, instance, args, kwargs):
    tracked_request = TrackedRequest.instance()
    camel_name = "".join(c.title() for c in wrapped.__name__.split("_"))
    operation = "MongoDB/{}.{}".format(instance.name, camel_name)
    span = tracked_request.start_span(operation=operation, ignore_children=True)
    span.tag("name", instance.name)

    try:
        return wrapped(*args, **kwargs)
    finally:
        tracked_request.stop_span()



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import os

from scout_apm.compat import kwargs_only
from scout_apm.core import objtrace
from scout_apm.core.config import scout_config
from scout_apm.core.core_agent_manager import CoreAgentManager
from scout_apm.core.instrument_manager import InstrumentManager
from scout_apm.core.metadata import report_app_metadata
from scout_apm.core.socket import CoreAgentSocket

logger = logging.getLogger(__name__)


@kwargs_only
def install(config=None):
    if config is not None:
        scout_config.set(**config)
    scout_config.log()

    if os.name == "nt":
        logger.info(
            "APM Not Launching on PID: %s - Windows is not supported", os.getpid()
        )
        return False

    if not scout_config.value("monitor"):
        logger.info(
            "APM Not Launching on PID: %s - Configuration 'monitor' is not true",
            os.getpid(),
        )
        return False

    InstrumentManager().install_all()

    objtrace.enable()

    logger.debug("APM Launching on PID: %s", os.getpid())
    launched = CoreAgentManager().launch()

    report_app_metadata()
    if launched:
        # Stop the thread to avoid running threads pre-fork
        CoreAgentSocket.instance().stop()

    return True



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import wrapt
from starlette.background import BackgroundTask
from starlette.requests import Request

import scout_apm.core
from scout_apm.core.tracked_request import TrackedRequest
from scout_apm.core.web_requests import (
    create_filtered_path,
    ignore_path,
    track_amazon_request_queue_time,
    track_request_queue_time,
)


class ScoutMiddleware:
    def __init__(self, app):
        self.app = app
        installed = scout_apm.core.install()
        self._do_nothing = not installed
        if installed:
            install_background_instrumentation()

    async def __call__(self, scope, receive, send):
        if self._do_nothing or scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        request = Request(scope)
        tracked_request = TrackedRequest.instance()
        # Can't name controller until post-routing - see final clause
        controller_span = tracked_request.start_span(operation="Controller/Unknown")

        tracked_request.tag(
            "path",
            create_filtered_path(request.url.path, request.query_params.multi_items()),
        )
        if ignore_path(request.url.path):
            tracked_request.tag("ignore_transaction", True)

        user_ip = (
            request.headers.get("x-forwarded-for", default="").split(",")[0]
            or request.headers.get("client-ip", default="").split(",")[0]
            or request.client.host
        )
        tracked_request.tag("user_ip", user_ip)

        queue_time = request.headers.get(
            "x-queue-start", default=""
        ) or request.headers.get("x-request-start", default="")
        tracked_queue_time = track_request_queue_time(queue_time, tracked_request)
        if not tracked_queue_time:
            amazon_queue_time = request.headers.get("x-amzn-trace-id", default="")
            track_amazon_request_queue_time(amazon_queue_time, tracked_request)

        def grab_extra_data():
            if "endpoint" in scope:
                # Rename top span
                endpoint = scope["endpoint"]
                controller_span.operation = "Controller/{}.{}".format(
                    endpoint.__module__, endpoint.__qualname__
                )
                tracked_request.is_real_request = True

            # From AuthenticationMiddleware - bypass request.user because it
            # throws AssertionError if 'user' is not in Scope, and we need a
            # try/except already
            try:
                username = scope["user"].display_name
            except (KeyError, AttributeError):
                pass
            else:
                tracked_request.tag("username", username)

        async def wrapped_send(data):
            # Finish HTTP span when body finishes sending, not later (e.g.
            # after background tasks)
            if data.get("type", None) == "http.response.body" and not data.get(
                "more_body", False
            ):
                grab_extra_data()
                tracked_request.stop_span()
            return await send(data)

        try:
            await self.app(scope, receive, wrapped_send)
        except Exception as exc:
            tracked_request.tag("error", "true")
            raise exc
        finally:
            if tracked_request.end_time is None:
                grab_extra_data()
                tracked_request.stop_span()


background_instrumentation_installed = False


def install_background_instrumentation():
    global background_instrumentation_installed
    if background_instrumentation_installed:
        return
    background_instrumentation_installed = True

    @wrapt.decorator
    async def wrapped_background_call(wrapped, instance, args, kwargs):
        tracked_request = TrackedRequest.instance()
        tracked_request.is_real_request = True
        tracked_request.start_span(
            operation="Job/{}.{}".format(
                instance.func.__module__, instance.func.__qualname__
            )
        )
        try:
            return await wrapped(*args, **kwargs)
        finally:
            tracked_request.stop_span()

    BackgroundTask.__call__ = wrapped_background_call(BackgroundTask.__call__)
